from typing import Dict
import re


class StopEval(Exception):
    pass


class Template:

    def __init__(self, template_name: str, source_name: str, raw_content: str, templum):
        self.template_name = template_name
        self.source_name = source_name
        self.target_name = ''
        self.raw_content = raw_content
        self._templum = templum

    def stop(self):
        raise StopEval()

    def evaluate(self, indent: int = 0, **kwargs):
        scope = {
            'self': self,
            'templum': self._templum
        }

        scope.update(kwargs)
        local_scope: Dict = {}

        def repl_exec(match) -> str:
            lines = [s for s in match.group(0).strip()[2:-2].split('\n')]
            for i, line in enumerate(lines):
                if line.startswith('\t'):
                    lines[i] = line[1:]
                elif line.startswith(' ' * 4):
                    lines[i] = line[4:]
            try:
                code = compile('\n'.join(lines), self.source_name, 'exec')
                exec(code, scope, local_scope)
                return ''
            except StopEval:
                return ''
            except Exception as e:
                raise Exception(f'exception in template {self.source_name}')

        def repl_eval(match) -> str:
            line = match.group(0)[2:-2].strip()
            try:
                code = compile(line, self.source_name, 'eval')
                return str(eval(code, scope, {}))
            except StopEval:
                return match.group(0)
            except Exception as e:
                raise Exception(f'exception in template {self.source_name}')

        content = self.raw_content
        content = re.sub(r'{#.*?#}([^\n]*\n)?', repl_exec, content, 0, re.S)
        scope.update(local_scope)
        content = re.sub('{%.*?%}', repl_eval, content, 0, re.S)

        if indent > 0:
            lines = content.splitlines(keepends=True)
            for i, line in enumerate(lines):
                if line.strip() != '':
                    lines[i] = '\t' + line
            content = ''.join(lines)

        return content
